#' Plot voltrations
#'
#' @param SI Stain index list as generated by estimate_brightness
#' @param optimal_voltage Rownames of the SI$SI data.frame to use
#' @param dummy_voltages Voltages to grey out in the overview
#' @param cofactor Arcsinh cofactor used for visualisation
#'
#' @importFrom rlang .data
#' @importFrom ggplot2 ggplot aes geom_point geom_segment geom_violin
#'                     ggtitle theme theme_minimal xlab ylab layer_scales
#'                     xlim geom_line scale_color_manual
#' @importFrom dplyr `%>%`
#' @importFrom patchwork plot_layout
#'
#' @export
plot_voltrations <- function(SI,
                             optimal_voltage = NULL,
                             dummy_voltages = NULL,
                             cofactor = 150) {
  cells <- SI$cells
  max_signal <- SI$max_signal
  if (is.null(optimal_voltage)) optimal_voltage <- SI$best_per_group
  SI <- SI$SI

  plots_individual <- list()
  for (group in unique(SI$Group)) {
    ids <- rownames(SI)[group == SI$Group]
    cell_subset <- cells %>%
      dplyr::filter(.data$ID %in% ids)

    p_cells <- ggplot(cell_subset) +
      geom_violin(
        aes(
          x = as.numeric(SI[.data$ID, "Voltage"]),
          y = asinh(.data$Value / cofactor),
          group = .data$ID
        ),
        fill = "black"
      ) +
      xlab("") +
      ylab("Expression") +
      theme_minimal() +
      theme(axis.text.x = element_text(
        angle = 90,
        vjust = 0.5,
        hjust = 1
      )) +
      ggtitle(group) +
      geom_point(
        aes(
          x = as.numeric(.data$Voltage),
          y = asinh(.data$MFI_pos / cofactor)
        ),
        col = "red",
        data = SI[ids, ]
      ) +
      geom_segment(
        aes(
          x = as.numeric(.data$Voltage) - 20,
          xend = as.numeric(.data$Voltage) + 20,
          y = asinh(.data$MFI_pos / cofactor),
          yend = asinh(.data$MFI_pos / cofactor)
        ),
        col = "red",
        data = SI[ids, ]
      ) +
      geom_point(
        aes(
          x = as.numeric(.data$Voltage),
          y = asinh(.data$MFI_neg / cofactor)
        ),
        col = "cyan",
        data = SI[ids, ]
      ) +
      geom_segment(
        aes(
          x = as.numeric(.data$Voltage) - 20,
          xend = as.numeric(.data$Voltage) + 20,
          y = asinh(.data$Max_neg / cofactor),
          yend = asinh(.data$Max_neg / cofactor)
        ),
        col = "cyan",
        data = SI[ids, ]
      ) +
      geom_segment(
        aes(
          x = as.numeric(.data$Voltage) - 20,
          xend = as.numeric(.data$Voltage) + 20,
          y = asinh(.data$Min_neg / cofactor),
          yend = asinh(.data$Min_neg / cofactor)
        ),
        col = "cyan",
        data = SI[ids, ]
      ) +
      ggplot2::geom_hline(yintercept = asinh(max_signal / cofactor))

    p_SI <- ggplot(
      SI[ids, ],
      aes(x = as.numeric(.data$Voltage), y = .data$SI)
    ) +
      geom_line() +
      geom_point(aes(col = as.character(.data$ID %in% optimal_voltage +
        2 * (.data$Pctg_OutOfRange > 0.01 |
          .data$Voltage %in% dummy_voltages)))) +
      theme_minimal() +
      scale_color_manual(values = c(
        "0" = "black",
        "1" = "red",
        "2" = "grey"
      )) +
      theme(legend.position = "none") +
      xlab("Voltage")

    p_rSD <- ggplot(
      SI[ids, ],
      aes(x = as.numeric(.data$Voltage), y = .data$rSD)
    ) +
      geom_line() +
      geom_point(aes(col = as.character(.data$ID %in% optimal_voltage +
        2 * (.data$Pctg_OutOfRange > 0.01 |
          .data$Voltage %in% dummy_voltages)))) +
      theme_minimal() +
      scale_color_manual(values = c(
        "0" = "black",
        "1" = "red",
        "2" = "grey"
      )) +
      theme(legend.position = "none") +
      xlab("Voltage")

    plots_individual[[group]] <-
      p_cells +
      (p_SI + xlim(layer_scales(p_cells)$x$range$range)) +
      (p_rSD + xlim(layer_scales(p_cells)$x$range$range)) +
      plot_layout(ncol = 1, heights = c(3, 1, 1))
  }

  return(plots_individual)
}
