#######################################################################
# Copyright (c) 2023 Juan Hern�ndez, FlowJo, LLC
#
# License
# The software is distributed under the terms of the 
# Artistic License 2.0
# http://www.r-project.org/Licenses/Artistic-2.0
# 
# Disclaimer
# This software and documentation come with no warranties of any kind.
# This software is provided "as is" and any express or implied 
# warranties, including, but not limited to, the implied warranties of
# merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the  copyright holder be liable for any direct, 
# indirect, incidental, special, exemplary, or consequential damages
# (including but not limited to, procurement of substitute goods or 
# services; loss of use, data or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort arising in any way out of the use of this 
# software.
######################################################################

## For debugging purposes
sessionInfo()

## For debugging to see what FlowJo is passing to the script.
##
## cat("FJ-SCRIPT-DIR			     = FJ_SCRIPT_DIR \n")
## cat("FJ-OUTPUT-DIR			     = FJ_OUTPUT_DIR \n")
## cat("FJ-FCS-FILES-PATH	     	 = FJ_FCS_FILES_PATH \n")
## cat("FJ-OUTPUT-RESULT-FILE          = FJ_OUTPUT_RESULT_FILE \n")
## cat("FJ-PREGATE      			 = FJ_PREGATE \n")
## cat("FJ-REMOVE-MARGINS            = FJ_REMOVE_MARGINS \n")
## cat("FJ-REMOVE-DOUBLETS           = FJ_REMOVE_DOUBLETS \n")
## cat("FJ-DETECTORS                  = FJ_DETECTORS \n")
## cat("FJ-FLUOROCHROMES              = FJ_FLUOROCHROMES \n")
## cat("FJ-REPORT              		 = FJ_REPORT \n")
## cat("FJ-FORMAT              		 = FJ_FORMAT \n")

# Install function for packages
loadPackages <-function(...) {
  libs<-unlist(list(...))
  req<-unlist(lapply(libs,require,character.only=TRUE))
  need<-libs[req==FALSE]
  if(length(need)>0){
    install.packages(need, repos="https://cran.r-project.org")
    lapply(need,require,character.only=TRUE)
  }
}

loadBioc<-function(...) {
  libs<-unlist(list(...))
  req<-unlist(lapply(libs,require,character.only=TRUE))
  need<-libs[req==FALSE]
  if(length(need)>0){
    BiocManager::install(need)
    lapply(need,require,character.only=TRUE)
  }
}

loadPackages(c('BiocManager', 'patchwork', 'readxl', 'ggplot2', 'dplyr', 'stats', 'rlang', 'stringr','utils','methods','data.table'))

loadBioc(c('Biobase','PeacoQC','flowDensity','flowCore'))


createFCS <- function(f){
  fcs_file <- paste(f,'FCS',sep='.')
  if(file.exists(fcs_file)){
  	return(fcs_file)
  }
  print(paste("Creating...", fcs_file))
  file_keywords <- paste(file, '.keywords', sep = '');
  dframe <- read.table(file_keywords,header=FALSE,sep='\t', col.names = c('Key', 'Value'))
  dtable <- data.table(dframe,key='Key')
  keywords <- list()
  for(k in dframe$Key) {
    keywords[[k]] <- dtable[k]$Value
  }
  
  dta = read.csv(f, header = TRUE, check.names = FALSE)
  dta<- as.matrix(dta)
  
  par <- length(dimnames(dta)[[2]])
  pnns <- c()
  pnss <- c()
  for(i in 1:par){
    pns <-paste('$P', i, 'S', sep = '')
    pnss[i] <- dtable[pns]$Value
  }

  vm <- data.frame(labelDescription = c(name = "Name of Parameter", 
                                        desc = "Description of Parameter", 
                                        range = "Range of Parameter", 
                                        minRange = "Minimum Parameter Value after Transformation", 
                                        maxRange = "Maximum Parameter Value after Transformation"))
  cols <- colnames(dta)
  pd <- data.frame(name = cols, 
                   desc = pnss,
                   range = apply(dta, 2, max, na.rm = TRUE), 
                   minRange = apply(dta, 2, min, na.rm = TRUE), 
                   maxRange = apply(dta, 2, max, na.rm = TRUE), 
                   row.names = paste0("$P", seq_along(cols)))
  parameters <- new("AnnotatedDataFrame", pd, vm)
  
  ff <- new("flowFrame",
            exprs=dta,
            parameters=parameters,
            description=keywords)
  write.FCS(ff,fcs_file)
  return(fcs_file)
}

setwd("FJ_SCRIPT_DIR")
pdf(file="FJ_REPORT")

source(file = "R/estimate_brightness.R")
source(file = "R/estimate_gate.R")
source(file = "R/estimate_SI.R")
source(file = "R/filter_plot.R")
source(file = "R/find_cutoff.R")
source(file = "R/find_plateau.R")
source(file = "R/plot_brightness.R")
source(file = "R/plot_voltration.R")
source(file = "R/prep_FCS.R")
source(file = "R/sample_cells.R")

getwd()

#variables
files <- FJ_FCS_FILES_PATH
detectors <- FJ_DETECTORS
fls <- FJ_FLUOROCHROMES
resultFile <- "FJ_OUTPUT_RESULT_FILE"
pregate <- FALSE
if(FJ_PREGATE == TRUE){
list(list(
      marker_values = list(
        "FSC-A" = c(
          min = 256,
          range_min = 512,
          max = 750,
          range_max = 512
        ),
        "SSC-A" = c(
          max = 512,
          range_max = 512
        )
      )
    ))
}

#script

if("FJ_FORMAT" == "TXT"){
  s <- 1
  for(set in files) {
    i <- 1
    for(file in set){
      files[[s]][[i]] <- createFCS(file)
      i <- i + 1
    }
    s <- s + 1
  }
  i <- 1
}

single_stain = data.frame()
stainID <- 0
i <- 1
for(set in files) {
	for(file in set){
  		single_stain = rbind(single_stain, list(stainID<-stainID+1,fls[i],detectors[i],file, detectors[i]))
	}
	i <- i + 1
}
colnames(single_stain) <- c("ID", "Fluorochrome","Detector","File", "Group") 
                            
print(single_stain)

preprocessing_parameters =
  list(
    compensate = FALSE,
    removeMargins = FJ_REMOVE_MARGINS,
    removeDoublets = FJ_REMOVE_DOUBLETS,
    pregate = FJ_PREGATE,
    pregate_tf = FALSE
  )

#```
bg = estimate_brightness(single_stains = single_stain,
                    preprocessing_parameters = preprocessing_parameters,
                    return_cells = TRUE,
                    seed = 1,
                    comp = NULL,
                    transform = FALSE)

#```
print(bg$pregating_plots)
print(bg$best_per_group)
#print(plot_brightness(bg$SI))
print(plot_voltrations(bg))
#```

# save the result to CSV
to.csv = data.frame()
for(i in single_stain$ID){
  to.csv <- rbind(to.csv, list(bg$SI$Detector[i], 
                               bg$SI$Fluorochrome[i], 
                               bg$SI$Voltage[i],
                               bg$SI$Cutoff[i],
                               bg$SI$SI[i],
                               bg$best_per_group[bg$SI$Detector[i]] == i,
                               bg$SI$File[i],
                               bg$SI$Min_neg[i],
                               bg$SI$Max_neg[i]
  ))
}
colnames(to.csv) <- c("Detector", "Fluorochrome", "Voltage", "Cutoff", "SI", "Best", "File", "Min_neg", "Max_neg")
print(to.csv)
write.csv(to.csv, file=resultFile, row.names=FALSE, quote=FALSE)
#```
dev.off()